/***************************************************************************
 *                                                                         *
 *                         Powersave Daemon                                *
 *                                                                         *
 *          Copyright (C) 2004,2005 SUSE Linux Products GmbH               *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at you   *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA                  *
 *                                                                         *
 ***************************************************************************/

#include <sys/wait.h>
#include <sys/time.h>

#include "cpufreq.h"
#include "event.h"
#include "config.h"

using namespace std;
using namespace __gnu_cxx;

string Event::_scripts_dir = PUB_SCRIPT_DIR;

Event::Event()
{
	_id = -1;
	_timeout = -1;
	_starttime = -1;
	_sleep_time = 0;
	_name = "default";
	_current_program_2_execute = _programs_2_execute.begin();
	setupInternalActions();
}

Event::Event(string n):_name(n)
{
	_id = -1;
	_timeout = -1;
	_starttime = -1;
	_sleep_time = 0;
	_current_program_2_execute = _programs_2_execute.begin();
	setupInternalActions();
	setActions2Execute("ignore");
}

Event::Event(const Event & e)
{
	_id = e._id;
	_starttime = e._starttime;
	_sleep_time = e._sleep_time;
	_timeout = e._timeout;
	_name = e._name;
	_programs_2_execute = e._programs_2_execute;
	_current_program_2_execute = _programs_2_execute.begin();
	setupInternalActions();
}

Event & Event::operator=(Event & e)
{
	_id = e._id;
	_starttime = e._starttime;
	_sleep_time = e._sleep_time;
	_timeout = e._timeout;
	_name = e._name;
	_programs_2_execute = e._programs_2_execute;
	_current_program_2_execute = _programs_2_execute.begin();
	setupInternalActions();
	return *this;
}

Event::~Event()
{
	if (!_programs_2_execute.empty())
		_programs_2_execute.clear();

	if (!_internalActions.empty())
		_internalActions.clear();
}

void Event::setupInternalActions()
{
	_internalActions.push_back("ignore");
	_internalActions.push_back("throttle");
	_internalActions.push_back("dethrottle");
	_internalActions.push_back("do_suspend_to_disk");
	_internalActions.push_back("do_suspend_to_ram");
	_internalActions.push_back("do_standby");
	_internalActions.push_back("suspend_to_disk");
	_internalActions.push_back("suspend_to_ram");
	_internalActions.push_back("standby");
	_internalActions.push_back("screen_saver");
	_internalActions.push_back("reread_cpu_capabilities");
}

string Event::name()
{
	return _name;
}

int Event::execute_program(PM_Interface * pm_object, string acpi_event_line)
{
	pid_t pid;
	int x = 0;
	int eventID = ID();

	// set the program to execute to the current one
	string exec_script_path;

	/* This event has no actions, needed to not segfault if empty */

	if (_programs_2_execute.empty())
		return 1;

	if (isInternalAction(*_current_program_2_execute)) {
		if (executeInternalAction(pm_object, *_current_program_2_execute) < 0) {
			pDebug(DBG_DIAG,
			       "Internal action %s returned an error",
			       (*_current_program_2_execute).c_str());
			string c = "internal action failed: ";
			c.append(*_current_program_2_execute);
			pm_object->checkScriptReturn(_id, 1, c.c_str());
			return 0;
		} else {
			pDebug(DBG_INFO,
			       "Internal action %s executed",
			       (*_current_program_2_execute).c_str());
			string c = "internal action successfull: ";
			c.append(*_current_program_2_execute);
			pm_object->checkScriptReturn(_id, 0, c.c_str());
		}
		return 1;
	}

	exec_script_path = _scripts_dir + "/" + *_current_program_2_execute;
	if (access(exec_script_path.c_str(), X_OK)) {
		pDebug(DBG_WARN,
		       "Could not execute program %s for event %s: %s",
		       exec_script_path.c_str(), _name.c_str(), strerror(errno));
		string c = "Cannot access or execute: ";
		c.append(*_current_program_2_execute);
		pm_object->checkScriptReturn(_id, 1, c.c_str());
		return 0;
	}

	pid = fork();
	//child
	if (pid == 0) {
		int argCount = 6;
		char *argv[argCount];

		signal(SIGHUP, SIG_DFL);
		signal(SIGTERM, SIG_DFL);
		signal(SIGINT, SIG_DFL);
		signal(SIGQUIT, SIG_DFL);
		signal(SIGPIPE, SIG_DFL);
		signal(SIGCHLD, SIG_DFL);
		// close all fds
		int fdlimit = sysconf(_SC_OPEN_MAX);

		if (nice(_proxy_nice_factor) == -1) {
			pDebug(DBG_WARN,
			       "Could not set nice factor for proxy execution: %s",
			       strerror(errno));
		}

		for (int fd = 0; fd < fdlimit; fd++)
			close(fd);

		for (x = 0; x < argCount; x++)
			argv[x] = NULL;
		x = 0;

		pDebug(DBG_INFO,
		       "Going to execute external action %s",
		       (*_current_program_2_execute).c_str());

		/* first give name of the program as parameter */
		argv[x] = strdup((*_current_program_2_execute).c_str());
		x++;

		/* second: give name of happening event */
		argv[x] = strdup(_name.c_str());
		x++;

		/* third is the current scheme file name */
		argv[x] = strdup(pm_object->_config->current_scheme->file_name.c_str());
		x++;

		/* forth is acpi line */
		if (acpi_event_line == "") {
			acpi_event_line = "empty";
		}
		argv[x] = strdup(acpi_event_line.c_str());
		x++;

		/* fifth is the event id */
		argv[x] = new char[5];
		sprintf(argv[x], "%d", eventID);
		x++;

		argv[x] = NULL;

		pDebug(DBG_INFO,
		       "Program %s will be executed for event %s. Param 1: %s,"
		       "Param 2: %s",
		       exec_script_path.c_str(), _name.c_str(), argv[1], argv[2]);

		// set environment variables for scripts
		setenv("SCRIPT_DIR", PUB_SCRIPT_DIR, 1);
		setenv("SYSCONF_DIR", CONFIG_DIR, 1);
		setenv("BIN_DIR", BIN_DIR, 1);

		execv(exec_script_path.c_str(), argv);

		pDebug(DBG_ERR,
		       "Came back from subprocess when trying to execute %s for "
		       "event %s; Error: %s ",
		       exec_script_path.c_str(), _name.c_str(), strerror(errno));
		_exit(127);
	} else if (pid == -1) {	// fork failed
		pDebug(DBG_ERR, "Could not fork process: %s", strerror(errno));
	} else {
		// parent
		// no need to wait for return of script because it returns
		// through the script_return_helper. So return to main loop and wait
		// for other events from acpi or sockets
	}
	return 1;
}

void Event::setActions2Execute(const string actions_line)
{

	string action;
	int x = actions_line.size();
	int y = 0;
	if (actions_line == "")
		return;

	_programs_2_execute.clear();

	while (x > 0 && isspace(actions_line[x - 1]))
		x--;
	y = x - 1;
	do {
		while (y > 0 && !isspace(actions_line[y])) {
			y--;
		}
		if (y != 0)
			y++;
		if (x > y) {
			action = actions_line.substr(y, x - y);
			_programs_2_execute.push_front(action);
		} else
			break;
		y--;
		while (y > 0 && isspace(actions_line[y]))
			y--;
		x = y + 1;
	} while (y > 0);
}

bool Event::isInternalAction(const string action)
{
	for (std::list< string >::iterator it = _internalActions.begin();
	     it != _internalActions.end(); ++it) {
		
		if (!action.compare(*it)) {
			return true;
		}
	}

	return false;
}

int Event::executeInternalAction(PM_Interface * pm_object, const string action)
{
	int ret = 1;

	pDebug(DBG_INFO, "Executing internal action: %s", action.c_str());

	if (!action.compare("ignore")) {
		/* ... */
	} else if (!action.compare("throttle")) {
		pm_object->throttle();
	} else if (!action.compare("dethrottle")) {
		pm_object->dethrottle();
	}
	/* really do the suspend/standby */
	else if (!action.compare("do_suspend_to_disk")) {
		pm_object->sleep_wrapper("suspend2disk");
		pm_object->executeScript(pm_object->_SLEEP_SCRIPT, "suspend2disk");
		pm_object->sleep_wrapper("suspend2disk");
	} else if (!action.compare("do_suspend_to_ram")) {
		pm_object->sleep_wrapper("suspend2ram");
		pm_object->suspend_to_ram();
		pm_object->sleep_wrapper("suspend2ram");
	} else if (!action.compare("do_standby")) {
		pm_object->sleep_wrapper("standby");
		pm_object->standby();
		pm_object->sleep_wrapper("standby");
	} else if (!action.compare("screen_saver")) {
		pm_object->screenlock(this->_name);
	}
	/* throw a suspend/standby event (will trigger other needed
	   actions for proper suspend/standby */
	else if (!action.compare("suspend_to_disk")) {
		pm_object->executeEvent("global.suspend2disk");
	} else if (!action.compare("suspend_to_ram")) {
		pm_object->executeEvent("global.suspend2ram");
	} else if (!action.compare("standby")) {
		pm_object->executeEvent("global.standby");
	} else if (!action.compare("reread_cpu_capabilities")) {
		if (CPUFreq_Interface::isSupported()) {
			CPUFreq_Interface::rereadFrequencies();
		}
	} else
		ret = -1;
	return ret;
}

/*
---------------------------------------------------------------------
   the following is needed for managing events
---------------------------------------------------------------------
*/

unsigned long Event::timeElapsed()
{
	struct timeval cur_time;
	struct timezone dummy;
	gettimeofday(&cur_time, &dummy);

	// time elapsed in milliseconds
	unsigned long time_elapsed = cur_time.tv_sec * 1000 + cur_time.tv_usec /
		1000 - _starttime - _sleep_time;

	return time_elapsed;
}

void Event::setTimeout(unsigned long m_seconds)
{
	_timeout = m_seconds;
}

unsigned long Event::getTimeout()
{
	return _timeout;
}

void Event::setID(int i)
{
	_id = i;
}

int Event::ID()
{
	return _id;
}

slist < string >::iterator Event::end()
{
	return _programs_2_execute.end();
}

void Event::start()
{
	struct timeval cur_time;
	struct timezone dummy;
	gettimeofday(&cur_time, &dummy);

	this->_starttime = cur_time.tv_sec * 1000 + cur_time.tv_usec / 1000;
}

void Event::increaseSleeptime(unsigned long time)
{
	_sleep_time += time;
}

/*
---------------------------------------------------------------------
   end of managing part
---------------------------------------------------------------------
*/

#ifdef CREATE_DEBUG_OUTPUT

ostream & operator<<(ostream & os, const Event & ev)
{
	copy(ev.programs_2_execute.begin(), ev.programs_2_execute.end(), ostream_iterator < string > (os));
	return os;
}

#endif
