/***************************************************************************
 *                                                                         *
 *                         Powersave Daemon                                *
 *                                                                         *
 *          Copyright (C) 2004,2005 SUSE Linux Products GmbH               *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at you   *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA                  *
 *                                                                         *
 ***************************************************************************/

#include "config.h"
#include <sys/wait.h>
#include <sys/time.h>

#include "cpufreq.h"

#include "event_management.h"
#include "main_loop.h"
#include "brightness.h"
#include "clientConnection.h"

string PM_Interface::_INT_SCRIPT_PATH = SCRIPT_DIR;
string PM_Interface::_SLEEP_SCRIPT = "do_acpi_sleep";

/* apm workaround variables for sleep modes ******************/
PSD_SLEEP_STATE_REQUESTS PM_Interface::_requested_sleep_state = PSD_NO_SLEEP_REQUEST;
bool PM_Interface::_sleep_triggered = false;
/* apm workaround variables for sleep modes ******************/

int PM_Interface::rereadConfig()
{
	int ret = _config->readConfFiles();

	/* reset/reread everything that could have been changed in
	   configs ... */
	if (CPUFreq_Interface::isSupported()) {
		CPUFreq_Interface::setConfigs(_config->current_scheme->CPU_HIGH_LIMIT,
					      _config->current_scheme->JUMP_CPU_FREQ_MAX_LIMIT,
					      _config->current_scheme->CPUFREQ_HYSTERESIS,
					      _config->current_scheme->CONSIDER_NICE);
	}
	activateSettings();
	return ret;
}


PM_Interface::PM_Interface(GeneralConfig * config_obj)
{
	pDebug(DBG_DEBUG, "Constructor PM_Interface");

	// initialise everything we have in the constructor !
	// acpi/apm specifc things are initialised in the
	// appropriate ACPI/APM_Interface constructor

	// **** I N I T   S E C T I O N *************************/

	_config = config_obj;
	// this should never happen, hopefully the term signal cleans up memory and stops app */
	if (!_config || !_config->current_scheme || !_config->AC_scheme || !_config->Battery_scheme) {
		pDebug(DBG_ERR, "Current, AC or Battery scheme is not set. Exit.");
		kill(getpid(), SIGTERM);
	}

	_supported_sleeping_states = getSupportedSleepStates();

	_cooling_mode_supported = true;
	_thermal_trip_points_supported = true;

	// this will be != NULL if cpufrequency scaling is supported
	// else the mode will be by default _DYNAMIC!
	//
	if (!_config->disable_CPU_freq) {
		CPUFreq_Interface::createCPUFreqObject(_config->CPUFREQ_CONTROL, _config->POLL_INTERVAL * 1000);
		if (CPUFreq_Interface::isSupported()) {
			CPUFreq_Interface::setConfigs(_config->current_scheme->CPU_HIGH_LIMIT,
						      _config->current_scheme->JUMP_CPU_FREQ_MAX_LIMIT,
						      _config->current_scheme->CPUFREQ_HYSTERESIS,
						      _config->current_scheme->CONSIDER_NICE);
		} else {
			pDebug(DBG_ERR, "Cannot create cpufreq objects");
		}
	}
#ifdef HAVE_LIBHAL
	// check for hal pid. We cannot check for hal sevice on dbus
	// because on booting, haldaemon is running but is not providing
	// any information. So an error would be thrown everytime.
	if (checkACPI() == ACPI && access("/var/run/hal/haldaemon.pid", F_OK < 0)) {
		pDebug(DBG_ERR, "Haldaemon not running. Battery or ac adapter information "
				"will not be available.");
	}
#endif

	// allocate memory for event management object
	_eM = new EventManagement();

	// create a new client connection object
	_cC = new ClientConnection();

	_brightness = Brightness::Probe();
	_brightness->Init();
}

PM_Interface::~PM_Interface()
{

	pDebug(DBG_DEBUG, "PM Destructor: delete allocated Objects");

	executeEvent("daemon.terminate");
	delete(_cC);
	delete(_eM);
	delete(_brightness);
	CPUFreq_Interface::destroy();
	delete _config;
}

int PM_Interface::start()
{
	// ********  trigger start event ****************/
	executeEvent("daemon.start");
	

	if (_battery.updateBatteryState(_cur_state.AC_STATE) < 0) {
		pDebug(DBG_INFO, "No battery available.");
	} else {
		pDebug(DBG_INFO, "There is at least one battery available.");
		// reset battery state, so the first checkBatteryChanges throws a battery
		// event if necessary
		_battery.resetState();
	}
	
	// Warning: when we start, the scheme should change from "nothing" to "active
	// scheme", a daemon.scheme.change event should be thrown and activateSettings
	// should be called. In reality, it changes from "Performance" to "active scheme",
	// which on AC is often "Performance" => no change => no event => no activate.
	// Not fatal, but we should fix it. In the meantime, we work around it here:
	activateSettings();
	executeEvent("daemon.scheme.change");
	
	/* this also sets current scheme, do not change order */
	updateACState();
	activateScheme(_cur_state.AC_STATE);
	
	// if the battery is already low, throw an appropriate event.
	checkBatteryStateChanges();
	
	return 1;
}

void PM_Interface::checkBatteryStateChanges()
{

	pDebug(DBG_DEBUG, "Check battery state");
	// only switch if a lower battery level is reached
	if (_battery.updateBatteryState(_cur_state.AC_STATE) == 1) {
		switch (_battery.batteryState()) {
		case BAT_NORM_STATE:
			executeEvent("battery.normal");
			pDebug(DBG_INFO, "Battery state changed to normal");
			break;
		case BAT_WARN_STATE:
			executeEvent("battery.warning");
			pDebug(DBG_INFO, "Battery state changed to warning");
			break;
		case BAT_LOW_STATE:
			executeEvent("battery.low");
			pDebug(DBG_INFO, "Battery state changed to low");
			break;
		case BAT_CRIT_STATE:
			executeEvent("battery.critical");
			pDebug(DBG_INFO, "Battery state changed to critical");
			break;
		case BAT_NONE_STATE:
			pDebug(DBG_INFO, "Battery disappeared.");
			break;
		default:
			pDebug(DBG_ERR, "Undefined battery state ..., that's bad");
		}
	} else {
		pDebug(DBG_INFO, "Battery state unchanged");
	}
}

int PM_Interface::checkACStateChanges()
{
	AC_ADAPTER_STATE ac_state;
	if (updateACState()) {
		pDebug(DBG_DIAG, "AC state changed");
		ac_state = _cur_state.AC_STATE;

		if (ac_state == AC_OFFLINE) {
			// set battery state
			executeEvent("acadapter.offline");
			pDebug(DBG_INFO, "ac adapter offline event occured");
			checkBatteryStateChanges();
			return 1;
		} else if (ac_state == AC_ONLINE) {
			executeEvent("acadapter.online");
			pDebug(DBG_INFO, "ac adapter online event occured");
			return 1;
		} else if (ac_state == AC_UNKNOWN) {
			// assume online for now...
			pDebug(DBG_DIAG, "AC state changed to unknown, assume online");
			ac_state = AC_ONLINE;
			return 1;
		} else {
			pDebug(DBG_ERR, "Could not read out AC adapter state, FATAL ERROR");
			return -1;
		}
	} else
		return 0;
}

void PM_Interface::checkTemperatureStateChanges()
{
	int ret = 0;
	pDebug(DBG_DEBUG, "Check temperature State");
	ret = updateTemperatureState();
	/* it is getting hotter */
	if (ret == 1) {
		switch (_cur_state.TEMP_STATE) {
		case CRITICAL:
			executeEvent("temperature.critical");
			pDebug(DBG_INFO, "Temperature state changed to critical.");
			break;
		case HOT:
			executeEvent("temperature.hot");
			pDebug(DBG_INFO, "Temperature state changed to hot.");
			break;
		case PASSIVE:
			executeEvent("temperature.passive");
			pDebug(DBG_INFO, "Temperature state changed to passive.");
			break;
			/* I do not differ between ACTIVE (fans on) and
			   OK mode (no cooling) at the moment */
		case ACTIVE:
			executeEvent("temperature.active");
			pDebug(DBG_INFO, "Temperature state changed to active.");
			break;
		case OK:
			pDebug(DBG_DIAG, "Temperature state changed to ok - this should"
			       "not happen at this point ->bug ->mail@renninger.de.");
			break;
		default:
			pDebug(DBG_DIAG, "Undefined temperature state ->bug ->mail@renninger.de.");
			break;
		}
	}
	/* it is getting cooler */
	else if (ret == 2) {
		switch (_cur_state.TEMP_STATE) {
		case CRITICAL:
			executeEvent("temperature.critical");
			pDebug(DBG_INFO, "Temperature state changed to critical. This should"
			       "not happen at this point ->bug ->mail@renninger.de.");
			break;
		case HOT:
			executeEvent("temperature.hot");
			pDebug(DBG_INFO, "Temperature state changed to hot.");
			break;
		case PASSIVE:
			executeEvent("temperature.passive");
			pDebug(DBG_INFO, "Temperature state changed to passive.");
			break;
			/* I do not differ between ACTIVE (fans on) and
			   OK mode (no cooling) at the moment */
		case ACTIVE:
			executeEvent("temperature.active");
			pDebug(DBG_INFO, "Temperature state changed to active.");
			break;
		case OK:
			executeEvent("temperature.ok");
			pDebug(DBG_INFO, "Temperature state changed to ok.");
			break;
		default:
			pDebug(DBG_INFO, "Undefined temperature state ->bug ->mail@renninger.de.");
			break;
		}
	}
}



bool PM_Interface::executeEvent(const string eventName, const string acpi_event_line)
{
	/* this is only for acpi events... */
	Event execEvent = _config->getEvent(eventName);
	int eventId;

	if (execEvent.name() == "global.suspend2disk" ||
	    execEvent.name() == "global.suspend2ram" ||
	    execEvent.name() == "global.standby") {

		if (_eM->eventIsRunning(execEvent.name())) {
			pDebug(DBG_WARN,
			       "Event %s already running. Ignoring this one",
			       execEvent.name().c_str());
			return true;
		}
	}

	if (execEvent.name() == "default") {
		pDebug(DBG_ERR, "Tried to execute non-existing event.");
		return false;
	}

	DBus_Server::emitSignal("PowersaveEvent", eventName.c_str());

	// append the event to the list of pending events. Timeout is set to 120 seconds.
	// are 2 minutes enough?
	eventId = _eM->registerEvent(execEvent, 120);

	// do not ignore child if daemon terminates
	// this is the only event where main process has
	// to wait until proxy is finished
	if (execEvent.name() == "daemon.terminate")
		signal(SIGCHLD, SIG_DFL);

	// execute the first script/internal action of the event
	execEvent.execute_program(this, acpi_event_line);

	return true;
}

bool PM_Interface::checkThrottling()
{
	return _throttleInterface.updateCPUState();
}

int PM_Interface::updateACState()
{
	int changed;
	int old_ac_state;

	old_ac_state = _cur_state.AC_STATE;
	_cur_state.AC_STATE = (AC_ADAPTER_STATE) getACAdapterStatus();

	/* there are new error codes returned 
	   (e.g. NO_DEVICE_ERROR and NO_MODULE_ERROR) */
	if (_cur_state.AC_STATE != AC_ONLINE && _cur_state.AC_STATE != AC_OFFLINE) {
		pDebug(DBG_INFO, "Could not detect AC status, assume AC_ONLINE");
		_cur_state.AC_STATE = AC_ONLINE;
	}
	if (old_ac_state != _cur_state.AC_STATE) {
		changed = 1;
		if (activateScheme(_cur_state.AC_STATE) > 0) {
			executeEvent("daemon.scheme.change");
		}
	} else
		changed = 0;
	return changed;
}

int PM_Interface::updateTemperatureState()
{

	int x;
	int state;
	/* is also THERM_MODE but C (library) forbids the declaration */
	int temp_state = OK;

	for (x = 0; _config->current_scheme->thermal_zones[x].present && (state = getThermalZoneState(x)) > 0 &&
	     x < MAX_THERMAL_ZONES; x++) {
		_config->current_scheme->thermal_zones[x].state = state;
		temp_state = max(state, temp_state);
	}
	/* here we need more proper error handling -> destroy allocated objects
	   probably the thermal module has been unloaded ... */
	if (_config->current_scheme->thermal_zones[x].present) {
		pDebug(DBG_DIAG, "Error during temperature evaluation");
		return -1;
	}
	if (temp_state > _cur_state.TEMP_STATE) {
		_cur_state.TEMP_STATE = (THERM_MODE) temp_state;
		return 1;
	} else if (temp_state < _cur_state.TEMP_STATE) {
		_cur_state.TEMP_STATE = (THERM_MODE) temp_state;
		return 2;
	} else
		return 0;
}

int PM_Interface::activateScheme(string scheme_name)
{

	int ret;
	if (!_config->current_scheme->SCHEME_NAME.compare(scheme_name)) {
		pDebug(DBG_DIAG, "Scheme already set");
		return 0;
	}
	ret = _config->setActiveScheme(scheme_name);
	if (ret <= 0)
		return ret;

	pDebug(DBG_DIAG, "switched scheme to: %s", scheme_name.c_str());
	activateSettings();
	return 1;
}

int PM_Interface::activateScheme(AC_ADAPTER_STATE ac_state)
{

	int ret = 0;
	switch (ac_state) {
	case AC_ONLINE:
		ret = activateScheme(_config->getACScheme());
		break;
	case AC_OFFLINE:
		ret = activateScheme(_config->getBatteryScheme());
		break;
	case AC_UNKNOWN:
		// which one should we use -> I'd say better use AC
		// we could even test for battery level, or is this
		// too overdosed?
		ret = activateScheme(_config->getACScheme());
		break;
	}
	return ret;
}

void PM_Interface::activateSettings()
{

	/***** experimental: Switch kernel/userspace control during runtime
	       We should be able to just override old objects as we do not use new
	       any more **********************************************************/
	if (_config->current_scheme->CPUFREQ_CONTROL != CPUFreq_Interface::getControlMode()) {
		pDebug(DBG_DIAG, "Change cpufreq control to: %s",
		       ((_config->current_scheme->CPUFREQ_CONTROL == CPUFREQ_USERSPACE) ? "userspace" : "kernel"));
		CPUFreq_Interface::createCPUFreqObject(_config->current_scheme->CPUFREQ_CONTROL,
						       _config->current_scheme->POLL_INTERVAL * 1000);
	}
	/*************************************************************************/

	/* set new CPU freq mode */
	if (!CPUFreq_Interface::setModes(_config->current_scheme->CPUFREQUENCY) < 0) {
		pDebug(DBG_WARN, "Could not set CPUFreq mode");
	}

	switch (_config->current_scheme->DISPLAY_BRIGHTNESS) {
	case -1:		// ignore
		break;
	case -2:		// low
		_brightness->Min();
		break;
	case -3:		// medium
		_brightness->Med();
		break;
	case -4:		// high
		_brightness->Max();
		break;
	default:		// level
		_brightness->Set(_config->current_scheme->DISPLAY_BRIGHTNESS);
		break;
	}

	mainLoop->updateCallbacks();
}

void PM_Interface::throttle(int percent)
{
	if (_cur_state.TEMP_STATE < PASSIVE) {
		_throttleInterface.throttle(percent);
	}
}

void PM_Interface::dethrottle()
{
	if (_cur_state.TEMP_STATE < PASSIVE) {
		_throttleInterface.dethrottle();
	}
}

void PM_Interface::resume()
{
	pDebug(DBG_DIAG, "Daemon reinitialises system states, resume finished");
	_requested_sleep_state = PSD_NO_SLEEP_REQUEST;
	// reset frequency which may have changed without noticing it
	CPUFreq_Interface::reinitSpeeds();
	checkACStateChanges();
	checkBatteryStateChanges();
}

void PM_Interface::checkEventTimeouts()
{
	// go through all events and check if one of them timed out
	std::list < Event >::iterator it;
	for (it = _eM->Events.begin(); it != _eM->Events.end(); it++) {

		if (it->timeElapsed() > it->getTimeout()) {
			pDebug(DBG_WARN, "Timeout for event %s with id %d reached, "
			       "removing", it->name().c_str(), it->ID());
			_eM->Events.erase(it);
			/* deleting the object may result in invokation
			   of a non existing event (it->timeElapsed() above)
			   -> start reiterating from the beginning of the list
			 */
			it = _eM->Events.begin();
		}
	}
}

void PM_Interface::checkScriptReturn(int id, int returnValue, string comment)
{

	// get the program name
	string prog = _eM->cur_prog(id);

	// if a program matching an event was found
	if (!prog.empty()) {
		pDebug(DBG_INFO, "SCRIPT returned: Name: '%s', Return value: %d, "
				 "Comment: '%s'", prog.c_str(), returnValue, comment.c_str());
	} else {
		pDebug(DBG_DIAG, "Unknown SCRIPT returned. Return value %d, text: '%s'", returnValue,
		       comment.c_str());
		// allow popup windows also from unknown processes?
		// if (returnValue !=2)
		return;
	}

	// enable the buttons when we come back from restore
	if (prog.find("restore_after_s") == 0) {
		ignore_buttons(false);
	}
	// evaluate the return code
	switch (returnValue) {
		// script succeeded
	case 0:{
			// set the event to the next script
			int i = _eM->moveOn(id);

			// no scripts left, so finished successfully
			if (!i) {
				const string name = _eM->name(id);
				pDebug(DBG_INFO, "SCRIPT Event %s finished successfully", name.c_str());
				// notify clients here
				_eM->remove(id);
			} else {
				_eM->event(id)->execute_program(this, "");
			}
			break;
		}
		// script requests a graphical notification from the daemon
	case 2:
		x_notification(comment);
		break;
		// script requests a xlock/screensaver from the daemon
	case 3:
		screenlock(comment);
		break;
	case 4:
		DBus_Server::emitSignal("Progress", comment.c_str());
		break;
		// script has failed
	case 1:
		const string event_name = _eM->name(id);
		pDebug(DBG_INFO, "SCRIPT %s failed to execute", event_name.c_str());

		// if the suspend/standby event has failed --> enable buttons
		if ((event_name.find("global.suspend") == 0)
		    || (event_name.find("global.standby") == 0)) {
			ignore_buttons(false);
		}
		// notify clients here
		// x_notification(comment);
		// remove the event from the list of pending events
		_eM->remove(id);
		break;
	}
}

void PM_Interface::sleep_wrapper(string event)
{
	static bool before_sleep = true;
	static struct timeval cur_time;
	static struct timezone dummy;

	if (before_sleep) {
		if (event == "suspend2disk") {
			if (_config->current_scheme->SUSPEND2DISK_DELAY > 0)
				sleep(_config->current_scheme->SUSPEND2DISK_DELAY);
		} else if (event == "suspend2ram") {
			if (_config->current_scheme->SUSPEND2RAM_DELAY > 0)
				sleep(_config->current_scheme->SUSPEND2RAM_DELAY);
		} else if (event == "standby") {
			if (_config->current_scheme->STANDBY_DELAY > 0)
				sleep(_config->current_scheme->STANDBY_DELAY);
		}
		ignore_buttons(true);

		// we used to sync here, but we already sync in userspace and the
		// kernel syncs again if necessary => this only slows things down
//              sync();

		pDebug(DBG_INFO, "Set machine into sleep mode");

		before_sleep = false;
		gettimeofday(&cur_time, &dummy);
	} else {
		// the time where we went to sleep in milliseconds
		unsigned long sleeptime = cur_time.tv_sec * 1000 + cur_time.tv_usec / 1000;

		gettimeofday(&cur_time, &dummy);

		// the time were we are back in milliseconds
		unsigned long wakeuptime = cur_time.tv_sec * 1000 + cur_time.tv_usec / 1000;

		// substract the time we slept from all events
		_eM->adjustTimeouts(wakeuptime - sleeptime);

		pDebug(DBG_INFO, "Back from %s inside sleep_wrapper", event.c_str());

		resume();
		before_sleep = true;
		// throw a resume event
		executeEvent("global.resume." + event);
	}
}

int PM_Interface::executeScript(string script, string param)
{
	int ret = -1;
	int pid;
	int status;
	char *argv[3];
	argv[0] = new char[script.size() + 1];
	argv[1] = new char[param.size() + 1];
	strcpy(argv[0], script.c_str());
	strcpy(argv[1], param.c_str());
	argv[2] = NULL;

	string full_path_to_script = this->_INT_SCRIPT_PATH + "/" + script;

	pid = fork();
	if (!pid) {
		signal(SIGHUP, SIG_DFL);
		signal(SIGTERM, SIG_DFL);
		signal(SIGINT, SIG_DFL);
		signal(SIGQUIT, SIG_DFL);
		signal(SIGPIPE, SIG_DFL);
		signal(SIGCHLD, SIG_DFL);
		// close all fds
		int fdlimit = sysconf(_SC_OPEN_MAX);

		for (int fd = 0; fd < fdlimit; fd++)
			close(fd);

		// set environment variables for scripts
		setenv("SCRIPT_DIR", SCRIPT_DIR, 1);
		setenv("SYSCONF_DIR", CONFIG_DIR, 1);
		setenv("BIN_DIR", BIN_DIR, 1);

		execv(full_path_to_script.c_str(), argv);

		// we should never get here.
		pDebug(DBG_ERR, "could not execv the sleep script '%s'. Argument: '%s', error: '%s'",
		       full_path_to_script.c_str(), argv[1], strerror(errno));
		_exit(127);
	} else if (pid == -1) {
		pDebug(DBG_WARN, "Could not fork for sleep script");
	} else {
		// if the script is the sleep script, wait for its return
		if (script == this->_SLEEP_SCRIPT) {
			if (waitpid(pid, &status, 0) < 0) {
				if (errno == ECHILD) {
					pDebug(DBG_DIAG, "Sleep script already exited.");
				} else {
					pDebug(DBG_ERR, "Sleep script failed - %s", strerror(errno));
				}
			} else if (WIFEXITED(status)) {
				if (WEXITSTATUS(status)) {
					pDebug(DBG_WARN,
					       "Program to trigger sleep (%s) failed and exited with status %d",
					       full_path_to_script.c_str(), WEXITSTATUS(status));
					ret = -1;
				} else {
					/* everything is alright, go on executing ... */
					pDebug(DBG_INFO, "Program %s to trigger sleep executed successfully",
					       full_path_to_script.c_str());
					ret = 0;
				}
			} else if (WIFSIGNALED(status)) {
				pDebug(DBG_DIAG, "Program %s exited on signal %d",
				       full_path_to_script.c_str(), WTERMSIG(status));
				ret = -1;
			}

		}
		// no need to wait for script to return
	}
	return ret;
}

int PM_Interface::screenlock(string param)
{

	DBus_Server::emitSignal("Screenlock", param.c_str());

	if (!_cC->checkClientCapabilities(CAPABILITY_SCREENLOCK)) {
		executeScript("do_screen_saver", param);
	}
	return 1;
}

int PM_Interface::x_notification(string param)
{
	DBus_Server::emitSignal("Notification", param.c_str());

	if (!_cC->checkClientCapabilities(CAPABILITY_NOTIFICATIONS)) {
		executeScript("do_x_notification", param);
	}
	return 1;
}

void PM_Interface::ignore_buttons(bool ignore)
{
	pDebug(DBG_DIAG, "ignore_buttons is set to: %d", ignore);
	_sleep_triggered = ignore;
}

DBUS_ERROR PM_Interface::handleDBusRequest(DBusMessage * msg, DBusMessage * reply, POWERSAVE_MESSAGE_TYPE type)
{

	string msg_member = dbus_message_get_member(msg);
	string sender = dbus_message_get_sender(msg);
	uint32_t serial = dbus_message_get_serial(msg);

	/** @todo: Debug only, throw out when working... */
	pDebug(DBG_INFO, "Received msg: '%s'", msg_member.c_str());
	pDebug(DBG_DEBUG, "Obj.path '%s' serial '%d' from '%s'", dbus_message_get_path(msg), serial, sender.c_str());

	switch (type) {
	case MANAGER_MESSAGE:
		return handleManagerMessage(msg);
		break;
	case REQUEST_MESSAGE:
		return handleRequestMessage(msg, reply);
		break;
	case ACTION_MESSAGE:
		return handleActionMessage(msg);
		break;
	default:
		/* Should never happen -> get rid of message? */
		pDebug(DBG_ERR, "Received invalid message");
		return REPLY_INVALID_INTERFACE;
	}

	return REPLY_GENERAL_ERROR;
}

DBUS_ERROR PM_Interface::handleManagerMessage(DBusMessage * msg)
{

	string msg_member = dbus_message_get_member(msg);
	string sender = dbus_message_get_sender(msg);

	char *dummy = "";

	pDebug(DBG_DIAG, "Handling manager request");
	if (msg_member == "Connect") {
		int capabilities;
		/* add client. No 'request string/modes' check so far */
		if (!dbusGetMessageInteger(msg, &capabilities, 0)) {
			_cC->addClient(sender.c_str(), capabilities);
			return REPLY_SUCCESS;
		} else
			return REPLY_INVALID_PARAM;
	} else if (msg_member == "Disconnect") {
		_cC->checkClientRemoval(sender.c_str());
		return REPLY_SUCCESS;
	} else if (msg_member == "EventReturn") {
		int event_id, request;

		// get the event id
		if (dbusGetMessageInteger(msg, &event_id, 0) < 0) {
			pDebug(DBG_WARN, "Script returned but we could not get the event id");
			return REPLY_INVALID_PARAM;
		}
		// get the so called mode ( notify, progress, true, false )
		if (!dbusGetMessageInteger(msg, &request, 1) < 0) {
			pDebug(DBG_WARN, "Script returned but we could not to get the request");
			return REPLY_INVALID_PARAM;
		}
		// get the message
		if (dbusGetMessageString(msg, &dummy, 0) < 0) {
			pDebug(DBG_INFO, "Script returned but we could not to get the message."
			       " Continuing anyway...");
			dummy = "";
		}

		pDebug(DBG_DEBUG, "Script returned: event id: %d, request: %d, message: '%s'", event_id, request, dummy);

		checkScriptReturn(event_id, request, dummy);

		return REPLY_SUCCESS;
	} else
		return REPLY_INVALID_METHOD;

	return REPLY_GENERAL_ERROR;
}

DBUS_ERROR PM_Interface::handleRequestMessage(DBusMessage * msg, DBusMessage * reply)
{

	string msg_member = dbus_message_get_member(msg);
	char *dummy = "";

	/** @todo Also reply REPLY_HW_NOT_SUPPORTED for suspend*/
	if (msg_member == "AllowedSuspendToDisk" || msg_member == "suspend_to_disk_allowed") {
		if (_config->current_scheme->USER_SUSPEND2DISK_DISABLED)
			return REPLY_DISABLED;
		else
			return REPLY_SUCCESS;
	} else if (msg_member == "AllowedSuspendToRam" || msg_member == "suspend_to_ram_allowed") {
		if (_config->current_scheme->USER_SUSPEND2RAM_DISABLED)
			return REPLY_DISABLED;
		else
			return REPLY_SUCCESS;
	} else if (msg_member == "AllowedStandby" || msg_member == "standby_allowed") {
		if (_config->current_scheme->USER_STANDBY_DISABLED)
			return REPLY_DISABLED;
		else
			return REPLY_SUCCESS;
	} else if (msg_member == "BatteryChargingState") {
		switch (_battery.chargingState()) {
		case CHARG_STATE_UNKNOWN:
			return REPLY_HW_NOT_SUPPORTED;
		case CHARG_STATE_CHARGING:
			dummy = "charging";
			break;
		case CHARG_STATE_DISCHARGING:
			dummy = "discharging";
			break;
		case (CHARG_STATE_CHARGING | CHARG_STATE_DISCHARGING):
			dummy = "charging discharging";
			break;
		default:
			return REPLY_INVALID_METHOD;
		}
		dbus_message_append_args(reply, DBUS_TYPE_STRING, &dummy, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	} else if (msg_member == "BatteryRemainingTime") {
		int time = _battery.remainingTime();
		if (time < 0)
			return REPLY_HW_NOT_SUPPORTED;
		else {
			dbus_message_append_args(reply, DBUS_TYPE_INT32, &time, DBUS_TYPE_INVALID);
			return REPLY_SUCCESS;
		}
	} else if (msg_member == "BatteryRemainingPercent") {
		int percent = _battery.remainingPercent();
		if (percent < 0)
			return REPLY_HW_NOT_SUPPORTED;
		else {
			dbus_message_append_args(reply, DBUS_TYPE_INT32, &percent, DBUS_TYPE_INVALID);
			return REPLY_SUCCESS;
		}
	} else if (msg_member == "BatteryRemainingChargingTime") {
		int time = _battery.remainingChargingTime();
		if (time < 0)
			return REPLY_HW_NOT_SUPPORTED;
		else {
			dbus_message_append_args(reply, DBUS_TYPE_INT32, &time, DBUS_TYPE_INVALID);
			return REPLY_SUCCESS;
		}
	} else if (msg_member == "BatteryState") {
		switch (_battery.batteryState()) {
		case BAT_NONE_STATE:
			return REPLY_HW_NOT_SUPPORTED;
			break;
		case BAT_NORM_STATE:
			dummy = "normal";
			break;
		case BAT_WARN_STATE:
			dummy = "warning";
			break;
		case BAT_LOW_STATE:
			dummy = "low";
			break;
		case BAT_CRIT_STATE:
			dummy = "critical";
			break;
		default:
			return REPLY_GENERAL_ERROR;
			break;
		}
		dbus_message_append_args(reply, DBUS_TYPE_STRING, &dummy, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	} else if (msg_member == "CpufreqPolicy") {
		const char *policy;
		if (!CPUFreq_Interface::isSupported())
			return REPLY_HW_NOT_SUPPORTED;

		switch (CPUFreq_Interface::getMode()) {
		case _PERFORMANCE:
			policy = "performance";
			break;
		case _POWERSAVE:
			policy = "powersave";
			break;
		case _DYNAMIC:
			policy = "dynamic";
			break;
		default:
			return REPLY_HW_NOT_SUPPORTED;
		}
		dbus_message_append_args(reply, DBUS_TYPE_STRING, &policy, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	} else if (msg_member == "AcPower") {
		switch (_cur_state.AC_STATE) {
		case AC_ONLINE:
			dummy = "on";
			break;
		case AC_OFFLINE:
			dummy = "off";
			break;
		case AC_UNKNOWN:
			dummy = "unknown";
			break;
		default:
			return REPLY_HW_NOT_SUPPORTED;
		}
		dbus_message_append_args(reply, DBUS_TYPE_STRING, &dummy, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;
	} else if (msg_member == "SchemesGet") {
		int active = -1;
		int battery = -1;
		int ac_power = -1;
		SchemeConfig *scheme;
		const char *scheme_name;

		for (int x = 0; x < MAX_SCHEMES; x++) {
			scheme = _config->getSchemeByNumber(x);
			if (scheme == NULL)
				break;
			/* only set \n if there will be another string to append */
			scheme_name = scheme->SCHEME_NAME.c_str();
			dbus_message_append_args(reply, DBUS_TYPE_STRING, &scheme_name, DBUS_TYPE_INVALID);
			if (scheme == _config->current_scheme)
				active = x;
			if (scheme == _config->Battery_scheme)
				battery = x;
			if (scheme == _config->AC_scheme)
				ac_power = x;
		}
		/*
		   pDebug (DBG_DIAG, "Active: %d, battery: %d, ac_power: %d",
		   battery, active, ac_power);
		 */

		/* order of appended INT32 for request "schemes":
		 * 0: no. of currently active   scheme
		 * 1: no. of           battery  scheme
		 * 2: no. of           ac_power scheme
		 * Use the values of these appended INT32 to determine which
		 * appended strings are active/battery/ac_power scheme
		 */
		if (active == -1 || battery == -1 || ac_power == -1) {
			pDebug(DBG_ERR, "Could not find active/battery/ac_power scheme %d/%d/%d",
			       active, battery, ac_power);
			return REPLY_GENERAL_ERROR;
		} else {
			dbus_message_append_args(reply,
						 DBUS_TYPE_INT32, &active,
						 DBUS_TYPE_INT32, &battery,
						 DBUS_TYPE_INT32, &ac_power, DBUS_TYPE_INVALID);
		}

		return REPLY_SUCCESS;
	} else if (msg_member == "BrightnessGet") {
		int brt = _brightness->Get();

		if (brt < 0) {
			return REPLY_HW_NOT_SUPPORTED;
		}

		dbus_message_append_args(reply, DBUS_TYPE_INT32, &brt, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;

	} else if (msg_member == "BrightnessGetPercent") {
		int brt = _brightness->GetPercent();

		if (brt < 0) {
			return REPLY_HW_NOT_SUPPORTED;
		}

		dbus_message_append_args(reply, DBUS_TYPE_INT32, &brt, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;

	} else if (msg_member == "BrightnessLevelsGet") {
		int brt = _brightness->GetLevels();

		if (brt < 0) {
			return REPLY_HW_NOT_SUPPORTED;
		}

		dbus_message_append_args(reply, DBUS_TYPE_INT32, &brt, DBUS_TYPE_INVALID);
		return REPLY_SUCCESS;

	} else if (msg_member == "supp_sleep_states") {

	} else
		return REPLY_INVALID_METHOD;

	return REPLY_GENERAL_ERROR;
}

DBUS_ERROR PM_Interface::handleActionMessage(DBusMessage * msg)
{

	string msg_member = dbus_message_get_member(msg);
	char *dummy = "";

	if (msg_member == "SuspendToDisk" || msg_member == "suspend_to_disk") {
		if (_config->current_scheme->USER_SUSPEND2DISK_DISABLED) {
			pDebug(DBG_DIAG, "Suspend to disk is disabled in config.");
			return REPLY_DISABLED;
		} else {
			pDebug(DBG_INFO, "Client has requested suspend2disk");
			_requested_sleep_state = PSD_SUSPEND2DISK;
			/* used for workaround for apm to remembemer triggered sleep state (not needed at
			   this point any more)**** */

			// executing event now triggers suspend2disk, using do_suspend_to_disk variable
			executeEvent("global.suspend2disk");
			return REPLY_SUCCESS;
		}
	} else if (msg_member == "SuspendToRam" || msg_member == "suspend_to_ram") {
		if (_config->current_scheme->USER_SUSPEND2RAM_DISABLED) {
			pDebug(DBG_DIAG, "Suspend2ram is disabled in config.");
			return REPLY_DISABLED;
		} else if (!(_supported_sleeping_states & APM_SUSPEND) && !(_supported_sleeping_states & ACPI_S3)) {
			pDebug(DBG_DIAG, "Suspend2ram is not supported");
			return REPLY_HW_NOT_SUPPORTED;
		} else {
			pDebug(DBG_INFO, "Client has requested suspend2ram");
			_requested_sleep_state = PSD_SUSPEND2RAM;
			// executing event now triggers suspend2ram, using do_suspend2ram variable!
			executeEvent("global.suspend2ram");
			//suspend2ram();
			return REPLY_SUCCESS;
		}
	} else if (msg_member == "Standby" || msg_member == "standby") {
		if (_config->current_scheme->USER_STANDBY_DISABLED) {
			pDebug(DBG_DIAG, "Standby is disabled in config.");
			return REPLY_DISABLED;
		} else if (!(_supported_sleeping_states & APM_STANDBY) && !(_supported_sleeping_states & ACPI_S1)) {
			pDebug(DBG_DIAG, "Standby is not supported");
			return REPLY_HW_NOT_SUPPORTED;
		} else {
			pDebug(DBG_INFO, "Client has requested standby");
			/* workaround for apm to remembemer triggered sleep state ***************
			   but defined for whole PM_Interface to avoid dynamic cast */
			_requested_sleep_state = PSD_STANDBY;
			/* workaround for apm to remembemer triggered sleep state ************** */
			// executing event now triggers suspend2ram, using do_suspend_to_ram variable!
			executeEvent("global.standby");
			//suspend2ram();
			return REPLY_SUCCESS;
		}
	} else if (msg_member == "CpufreqPerformance") {
		if (CPUFreq_Interface::isSupported()) {
			if (CPUFreq_Interface::getMode() != _PERFORMANCE) {
				if (CPUFreq_Interface::setModes(_PERFORMANCE) < 0) {
					pDebug(DBG_ERR, "Could not set performance mode");
					return REPLY_HW_NOT_SUPPORTED;
				}
				_throttleInterface.restart();
				executeEvent("processor.performance");

				return REPLY_SUCCESS;
			} else
				return REPLY_ALREADY_SET;
		} else
			return REPLY_HW_NOT_SUPPORTED;
	} else if (msg_member == "CpufreqPowersave") {
		if (CPUFreq_Interface::isSupported()) {
			if (CPUFreq_Interface::getMode() != _POWERSAVE) {
				if (CPUFreq_Interface::setModes(_POWERSAVE) < 0) {
					pDebug(DBG_ERR, "Could not set powersave mode");
					return REPLY_HW_NOT_SUPPORTED;
				}
				// this only concerns dynamic events
				_throttleInterface.restart();
				executeEvent("processor.powersave");

				return REPLY_SUCCESS;
			} else
				return REPLY_ALREADY_SET;
		} else
			return REPLY_HW_NOT_SUPPORTED;
	} else if (msg_member == "CpufreqDynamic") {
		if (CPUFreq_Interface::isSupported()) {
			if (CPUFreq_Interface::getMode() != _DYNAMIC) {
				if (CPUFreq_Interface::setModes(_DYNAMIC) < 0) {
					pDebug(DBG_ERR, "Could not set dynamic mode");
					return REPLY_HW_NOT_SUPPORTED;
				}
				// this only concerns dynamic events
				_throttleInterface.restart();
				executeEvent("processor.dynamic");

				return REPLY_SUCCESS;
			} else
				return REPLY_ALREADY_SET;
		} else
			return REPLY_HW_NOT_SUPPORTED;
	} else if (msg_member == "Ping") {
		return REPLY_SUCCESS;
	} else if (msg_member == "SchemesSet") {
		if (!dbusGetMessageString(msg, &dummy, 0)) {
			int r = activateScheme(dummy);
			if (r == 0) {
				return REPLY_ALREADY_SET;
			} else if (r < 0) {
				pDebug(DBG_DIAG, "Scheme '%s' does not exist", dummy);
				return REPLY_INVALID_PARAM;
			} else {
				pDebug(DBG_DIAG, "Scheme '%s' received and activated", dummy);

				executeEvent("daemon.scheme.change");
				return REPLY_SUCCESS;
			}
		} else
			return REPLY_INVALID_PARAM;
	} else if (msg_member == "BrightnessSet") {
		int value;
		if (!dbusGetMessageInteger(msg, &value, 0)) {
			_brightness->Set(value);
			return REPLY_SUCCESS;
		} else {
			return REPLY_GENERAL_ERROR;
		}
	} else if (msg_member == "BrightnessSetPercent") {
		int value;
		if (!dbusGetMessageInteger(msg, &value, 0)) {
			_brightness->SetPercent(value);
			return REPLY_SUCCESS;
		} else {
			return REPLY_GENERAL_ERROR;
		}
	} else if (msg_member == "BrightnessMin") {
		_brightness->Min();
		return REPLY_SUCCESS;
	} else if (msg_member == "BrightnessMed") {
		_brightness->Med();
		return REPLY_SUCCESS;
	} else if (msg_member == "BrightnessMax") {
		_brightness->Max();
		return REPLY_SUCCESS;
	} else if (msg_member == "BrightnessUp") {
		_brightness->Up();
		return REPLY_SUCCESS;
	} else if (msg_member == "BrightnessDown") {
		_brightness->Down();
		return REPLY_SUCCESS;
	} else {
		return REPLY_INVALID_METHOD;
	}

	return REPLY_GENERAL_ERROR;
}

BATTERY_STATES PM_Interface::batteryState() {
	return _battery.batteryState();
}
